function perf = sdo_n_reps ( data, outlier, seed, runs, k, q, x )
% This script runs SDO a number of times represented by 'runs' and
% calculates performance indices.
% Dec 2017
%
% Requeriments:
% SDO.v1 algorithm (MATLAB)
%
% Inputs:
% 'data': MxN-matrix with M samples and N features/dimensions
% 'outlier': array of M samples, 0-inlier, 1-outlier
% 'seed': random seed basis, every experiment increases in one the current seed
% 'runs': number of different experiments
% 'k','q','x': SDO parameters (see "sdo.m")
%
% Outputs:
% 'perf.indices' contains the performance indices (columns) for every test (rows)
%           the meaning of the indices can be consulted in:
%           On the Evaluation of Unsupervised Outlier Detection: Measures, Datasets, and an Empirical Study
%           by G. O. Campos, A. Zimek, J. Sander, R. J. G. B. Campello, B. Micenková, E. Schubert, I. Assent and M. E. Houle
%           Data Mining and Knowledge Discovery 30(4): 891-927, 2016, DOI: 10.1007/s10618-015-0444-8
%           http://www.dbs.ifi.lmu.de/research/outlier-evaluation/DAMI/
% 'perf.parameters' contains the parameters (columns) for every test (rows)
%           1st col, random seed
%           2nd col, number of samples
%           3rd col, number of features
%           4th col, number of total SDO observers, i.e., 'k'
%           5th col, number of active SDO observers, i.e., 'kact'
%           6th col, closest observers (SDO parameter), i.e., 'x'
%           7th col, SDO observation threshold, i.e., 'q'
%           8th col, SDO quantile threshold for 'q' calculation, i.e., 'qv'
% 'perf.best_indices', the best experiment according to the obtained indices.
%   Indices are z-scored and summed up to calculate the best one 
% 'perf.best_parameters', parameters corresponding to the best experiment

if exist('seed')==0, seed=0;end
if exist('runs')==0, runs=10;end
if exist('x'), pi.x=x;end
if exist('k'), pi.k=k;end
if exist('q'), pi.q=q;end

for i=1:runs
    pi.sd=seed+i;
    fprintf('[%d] random seed: %d, SDO, ',i,pi.sd);
    [y,ob,po] = sdo(data, pi);
    fprintf('performance\n');
    index=[];
    [ index ] = performance ( y, outlier );
    indices(i,:)=[index.P_at_n, index.adjP_at_n, index.AP, index.adjAP, index.MaxF1, index.adj_MF1, index.ROC_AUC];
    indices_short(i,:)=[index.P_at_n, index.AP, index.MaxF1, index.ROC_AUC];
    parameters(i,:)=[po.sd, po.m, po.n, po.k, po.kp, po.x, po.q, po.qv];
end
z_ind=zscore(indices_short);
best_ind=sum(z_ind');
[val pos]=max(best_ind);
perf.best_indices=indices(pos,:);
perf.best_parameters=parameters(pos,:);
perf.indices=indices;
perf.parameters=parameters;
end
